@@ -1,3 +1,4 @@
+from array import array
 from datetime import datetime
 import logging
 import math
@@ -11,9 +12,7 @@ UNKNOWN = 1
 UNWALKABLE = 2
 WALKABLE = 3
 
-
-
-class Pathfinder:
+class Pathfinder(object):
     G = None
     FIRST_MAP = 'main'
     TRANSPORT_COST = 50
@@ -200,7 +199,8 @@ class Pathfinder:
         current = path[0]
         for i in range(1, len(path)):
             next = path[i]
-            cost += Pathfinder.computeLinkCost(current, next, next, avoidTownWarps=avoidTownWarps, costs=costs)
+            link = { 'data': { **next } }
+            cost += Pathfinder.computeLinkCost(current, next, link, avoidTownWarps=avoidTownWarps, costs=costs)
             current = next
         return cost
     
@@ -220,61 +220,55 @@ class Pathfinder:
         width = maxX - minX
         height = maxY - minY
 
+        # gridStart = datetime.utcnow().timestamp()
+        # print("  Starting grid creation...")
         grid = [UNKNOWN] * (height * width)
         for yLine in gGeo['y_lines']:
             lowerY = max([0, yLine[0] - minY - base['vn']])
-            upperY = yLine[0] - minY + base['v']
-            for y in range(lowerY, height):
-                if y > upperY:
-                    break
+            upperY = min([yLine[0] - minY + base['v'] + 1, height])
+            for y in range(lowerY, upperY):
                 lowerX = max([0, yLine[1] - minX - base['h']])
-                upperX = yLine[2] - minX + base['h']
-                for x in range(lowerX, width):
-                    if x > upperX:
-                        break
+                upperX = min([yLine[2] - minX + base['h'] + 1, width])
+                for x in range(lowerX, upperX):
                     grid[y * width + x] = UNWALKABLE
         
         for xLine in gGeo['x_lines']:
             lowerX = max([0, xLine[0] - minX - base['h']])
-            upperX = xLine[0] - minX + base['h']
-            for x in range(lowerX, width):
-                if x > upperX:
-                    break
+            upperX = min([xLine[0] - minX + base['h'] + 1, width])
+            for x in range(lowerX, upperX):
                 lowerY = max([0, xLine[1] - minY - base['vn']])
-                upperY = xLine[2] - minY + base['v']
-                for y in range(lowerY, height):
-                    if y > upperY:
-                        break
+                upperY = min([xLine[2] - minY + base['v'] + 1, height])
+                for y in range(lowerY, upperY):
                     grid[y * width + x] = UNWALKABLE
         
         for spawn in gMap['spawns']:
             x = math.trunc(spawn[0]) - minX
             y = math.trunc(spawn[1]) - minY
-            if grid[y * width + x] == WALKABLE:
-                continue
-            stack = [[y,x]]
-            while len(stack) > 0:
-                [y,x] = stack.pop()
-                while x >= 0 and grid[y * width + x] == UNKNOWN:
-                    x -= 1
-                x += 1
-                spanAbove = 0
-                spanBelow = 0
-                while x < width and grid[y * width + x] == UNKNOWN:
-                    grid[y * width + x] = WALKABLE
-                    if not spanAbove and y > 0 and grid[(y - 1) * width + x] == UNKNOWN:
-                        stack.append([y - 1, x])
-                        spanAbove = 1
-                    elif spanAbove and y > 0 and grid[(y - 1) * width + x] != UNKNOWN:
-                        spanAbove = 0
-                    
-                    if not spanBelow and y < (height - 1) and grid[(y + 1) * width + x] == UNKNOWN:
-                        stack.append([y + 1, x])
-                        spanBelow = 1
-                    elif spanBelow and y < (height - 1) and grid[(y + 1) * width + x] != UNKNOWN:
-                        spanBelow = 0
+            if grid[y * width + x] != WALKABLE:
+                stack = [[y,x]]
+                while len(stack) > 0:
+                    [y,x] = stack.pop()
+                    while x >= 0 and grid[y * width + x] == UNKNOWN:
+                        x -= 1
                     x += 1
-        
+                    spanAbove = 0
+                    spanBelow = 0
+                    while x < width and grid[y * width + x] == UNKNOWN:
+                        grid[y * width + x] = WALKABLE
+                        if not spanAbove and y > 0 and grid[(y - 1) * width + x] == UNKNOWN:
+                            stack.append([y - 1, x])
+                            spanAbove = 1
+                        elif spanAbove and y > 0 and grid[(y - 1) * width + x] != UNKNOWN:
+                            spanAbove = 0
+                        
+                        if not spanBelow and y < (height - 1) and grid[(y + 1) * width + x] == UNKNOWN:
+                            stack.append([y + 1, x])
+                            spanBelow = 1
+                        elif spanBelow and y < (height - 1) and grid[(y + 1) * width + x] != UNKNOWN:
+                            spanBelow = 0
+                        x += 1
+        # print(f"  grid size: {len(grid)}")
+        # print(f"  grid completion: {(datetime.utcnow().timestamp() - gridStart)}\n")
         Pathfinder.grids[map] = grid
         Pathfinder.updateGraph(grid, map)
         return grid
@@ -290,16 +284,15 @@ class Pathfinder:
         width = maxX - minX
         height = maxY - minY
 
-        walkableNodes = []  # list of walkable nodes
-        points = []         # list of points for delaunay
+        points = set()         # list of points for delaunay
         links = []          # list of links between walkable nodes as [point1, point2]
         linkData = []       # list of dictionaries containing attributes for particular links
         linkAttr = {}       # dictionary containing list of attributes for each link
 
         # add nodes at corners
-        cornersStart = datetime.now()
-        Pathfinder.logger.debug("  Adding corners...")
-        Pathfinder.logger.debug(f"  # nodes: {len(walkableNodes)}")
+        # cornersStart = datetime.utcnow().timestamp()
+        # print("  Adding corners...")
+        # print(f"  # nodes: {len(walkableNodes)}")
         for y in range(1, height - 1):
             for x in range(1, width):
                 # for each walkable x/y position...
@@ -320,54 +313,27 @@ class Pathfinder:
                 mapX = x + minX
                 mapY = y + minY
 
-                # inside-1
-                if (bottomLeft == UNWALKABLE) and (bottomCenter == UNWALKABLE) and (bottomRight == UNWALKABLE) and (middleLeft == UNWALKABLE) and (upperLeft == UNWALKABLE):
-                    walkableNodes.append(Pathfinder.addNodeToGraph(map, mapX, mapY))
-                    points.append([mapX, mapY])
-                # inside-2
-                elif (bottomLeft == UNWALKABLE) and (bottomCenter == UNWALKABLE) and (bottomRight == UNWALKABLE) and (middleRight == UNWALKABLE) and (upperRight == UNWALKABLE):
-                    walkableNodes.append(Pathfinder.addNodeToGraph(map, mapX, mapY))
-                    points.append([mapX, mapY])
-                # inside-3
-                elif (bottomRight == UNWALKABLE) and (middleRight == UNWALKABLE) and (upperLeft == UNWALKABLE) and (upperCenter == UNWALKABLE) and (upperRight == UNWALKABLE):
-                    walkableNodes.append(Pathfinder.addNodeToGraph(map, mapX, mapY))
-                    points.append([mapX, mapY])
-                # inside-4
-                elif (bottomLeft == UNWALKABLE) and (middleLeft == UNWALKABLE) and (upperLeft == UNWALKABLE) and (upperCenter == UNWALKABLE) and (upperRight == UNWALKABLE):
-                    walkableNodes.append(Pathfinder.addNodeToGraph(map, mapX, mapY))
-                    points.append([mapX, mapY])
-                # outside-1 
-                elif (bottomLeft == UNWALKABLE) and (bottomCenter == WALKABLE) and (middleLeft == WALKABLE):
-                    walkableNodes.append(Pathfinder.addNodeToGraph(map, mapX, mapY))
-                    points.append([mapX, mapY])
-                # outside-2
-                elif (bottomCenter == WALKABLE) and (bottomRight == UNWALKABLE) and (middleRight == WALKABLE):
-                    walkableNodes.append(Pathfinder.addNodeToGraph(map, mapX, mapY))
-                    points.append([mapX, mapY])
-                # outside-3
-                elif (middleRight == WALKABLE) and (upperCenter == WALKABLE) and (upperRight == UNWALKABLE):
-                    walkableNodes.append(Pathfinder.addNodeToGraph(map, mapX, mapY))
-                    points.append([mapX, mapY])
-                # outside-4
-                elif (middleLeft == WALKABLE) and (upperLeft == UNWALKABLE) and (upperCenter == WALKABLE):
-                    walkableNodes.append(Pathfinder.addNodeToGraph(map, mapX, mapY))
-                    points.append([mapX, mapY])
-        Pathfinder.logger.debug(f"  corner completion: {(datetime.now() - cornersStart).total_seconds()}\n")
+                # check corner locations for walkability
+                if (((WALKABLE not in [upperLeft, middleLeft, bottomLeft, bottomCenter, bottomRight]))      # inside-1
+                    or ((WALKABLE not in [upperRight, middleRight, bottomLeft, bottomCenter, bottomRight])) # inside-2
+                    or ((WALKABLE not in [upperLeft, upperCenter, upperRight, middleRight, bottomRight]))   # inside-3
+                    or ((WALKABLE not in [upperLeft, upperCenter, upperRight, middleLeft, bottomLeft]))     # inside-4
+                    or ((bottomLeft == UNWALKABLE) and (UNWALKABLE not in [middleLeft, bottomCenter]))      # outside-1
+                    or ((bottomRight == UNWALKABLE) and (UNWALKABLE not in [middleRight, bottomCenter]))    # outside-2
+                    or ((upperRight == UNWALKABLE) and (UNWALKABLE not in [upperCenter, middleRight]))      # outside-3
+                    or ((upperLeft == UNWALKABLE) and (UNWALKABLE not in [upperCenter, middleLeft]))):      # outside-4
+                    points.add((mapX, mapY))
+        # print(f"  corner completion: {(datetime.utcnow().timestamp() - cornersStart)}\n")
         
         # add nodes at transporters (we'll look for close nodes to transporters later)
-        transportersStart = datetime.now()
-        Pathfinder.logger.debug("  Adding transporter nodes...")
-        Pathfinder.logger.debug(f"  # nodes: {len(walkableNodes)}")
-        transporters = []
-        for npc in gMap['npcs']:
-            if npc['id'] != 'transporter':
-                continue
+        # transportersStart = datetime.utcnow().timestamp()
+        # print("  Adding transporter nodes...")
+        # print(f"  # nodes: {len(walkableNodes)}")
+        transporters = [npc for npc in gMap['npcs'] if npc['id'] == 'transporter']
+        for npc in transporters:
             pos = npc['position']
             closest = Pathfinder.findClosestSpawn(map, pos[0], pos[1])
-            fromNode = Pathfinder.addNodeToGraph(map, closest['x'], closest['y'])
-            points.append([closest['x'], closest['y']])
-            walkableNodes.append(fromNode)
-            transporters.append(npc)
+            points.add((closest['x'], closest['y']))
 
             # make more points around transporter
             angle = 0
@@ -375,28 +341,19 @@ class Pathfinder:
                 x = math.trunc(pos[0] + math.cos(angle) * (Constants.TRANSPORTER_REACH_DISTANCE - 10))
                 y = math.trunc(pos[1] + math.sin(angle) * (Constants.TRANSPORTER_REACH_DISTANCE - 10))
                 if Pathfinder.canStand({'map': map, 'x': x, 'y': y}):
-                    fromNode = Pathfinder.addNodeToGraph(map, x, y)
-                    points.append([x, y])
-                    walkableNodes.append(fromNode)
+                    points.add((x, y))
                 angle += math.pi / 32
-        Pathfinder.logger.debug(f"  transporter completion: {(datetime.now() - transportersStart).total_seconds()}\n")
+        # print(f"  transporter completion: {(datetime.utcnow().timestamp() - transportersStart)}\n")
 
         # add nodes at doors (we'll look for close nodes to doors later)
-        doorsStart = datetime.now()
-        Pathfinder.logger.debug("  Adding door nodes...")
-        Pathfinder.logger.debug(f"  # nodes: {len(walkableNodes)}")
-        doors = []
-        for door in gMap['doors']:
-            #TODO: Figure out how to know if we have access to a locked door
-            if len(door) > 7 and door[7] == 'complicated':
-                continue
-
+        # doorsStart = datetime.utcnow().timestamp()
+        # print("  Adding door nodes...")
+        # print(f"  # nodes: {len(walkableNodes)}")
+        doors = [door for door in gMap['doors'] if len(door) <= 7 or door[7] != 'complicated']
+        for door in doors:
             # From
             spawn = gMap['spawns'][door[6]]
-            fromDoor = Pathfinder.addNodeToGraph(map, spawn[0], spawn[1])
-            points.append([spawn[0], spawn[1]])
-            walkableNodes.append(fromDoor)
-            doors.append(door)
+            points.add((spawn[0], spawn[1]))
 
             # make more points around the door
             doorX = door[0]
@@ -415,31 +372,28 @@ class Pathfinder:
                     x = math.trunc(point['x'] + math.cos(angle) * (Constants.DOOR_REACH_DISTANCE - 10))
                     y = math.trunc(point['y'] + math.sin(angle) * (Constants.DOOR_REACH_DISTANCE - 10))
                     if Pathfinder.canStand({ 'map': map, 'x': x, 'y': y }):
-                        fromNode = Pathfinder.addNodeToGraph(map, x, y)
-                        points.append([x, y])
-                        walkableNodes.append(fromNode)
+                        points.add((x, y))
                     angle += math.pi / 32
-        Pathfinder.logger.debug(f"  door completion: {(datetime.now() - doorsStart).total_seconds()}\n")
+        # print(f"  door completion: {(datetime.utcnow().timestamp() - doorsStart)}\n")
         
         # Add nodes at spawns
-        spawnsStart = datetime.now()
-        Pathfinder.logger.debug("  Adding spawn nodes...")
-        Pathfinder.logger.debug(f"  # nodes: {len(walkableNodes)}")
-        townNode = Pathfinder.addNodeToGraph(map, gMap['spawns'][0][0], gMap['spawns'][0][1])
-        walkableNodes.append(townNode)
-        points.append([townNode['x'], townNode['y']])
-        townLinkData = { 'map': map, 'type': 'town', 'x': townNode['x'], 'y': townNode['y'], 'spawn': None, 'key': None }
-        for i in range(1, len(gMap['spawns'])):
-            spawn = gMap['spawns'][i]
-            node = Pathfinder.addNodeToGraph(map, spawn[0], spawn[1])
-            walkableNodes.append(node)
-            points.append([node['x'], node['y']])
-        Pathfinder.logger.debug(f"  spawns completion: {(datetime.now() - spawnsStart).total_seconds()}\n")
+        # spawnsStart = datetime.utcnow().timestamp()
+        # print("  Adding spawn nodes...")
+        # print(f"  # nodes: {len(walkableNodes)}")
+        townLinkData = { 'map': map, 'type': 'town', 'x': gMap['spawns'][0][0], 'y': gMap['spawns'][0][1], 'spawn': None, 'key': None }
+        for spawn in gMap['spawns']:
+            points.add((spawn[0], spawn[1]))
+        # print(f"  spawns completion: {(datetime.utcnow().timestamp() - spawnsStart)}\n")
+
+        vertexNames = [f"{map}:{x},{y}" for (x, y) in points]
+        attrs = { 'map': [map] * len(vertexNames), 'x': [p[0] for p in points], 'y': [p[1] for p in points] }
+        Pathfinder.graph.add_vertices(vertexNames, attrs)
+        walkableNodes = Pathfinder.graph.vs.select(map_eq=map)
         
         # collect link data
-        linksStart = datetime.now()
-        Pathfinder.logger.debug("  Adding walkable links...")
-        Pathfinder.logger.debug(f"  # nodes: {len(walkableNodes)}")
+        # linksStart = datetime.utcnow().timestamp()
+        # print("  Adding walkable links...")
+        # print(f"  # nodes: {len(walkableNodes)}")
         for fromNode in walkableNodes:
             # add destination nodes and links to maps that are reachable through the door(s)
             for door in doors:
@@ -474,6 +428,7 @@ class Pathfinder:
         
         leaveLink = Pathfinder.addNodeToGraph('main', Pathfinder.G['maps']['main']['spawns'][0][0], Pathfinder.G['maps']['main']['spawns'][0][1])
         leaveLinkData = { 'map': leaveLink['map'], 'type': 'leave', 'x': leaveLink['x'], 'y': leaveLink['y'], 'key': None, 'spawn': None }
+        townNode = Pathfinder.graph.vs.find(name=f"{map}:{gMap['spawns'][0][0]},{gMap['spawns'][0][1]}")
         for node in walkableNodes:
             # create town links
             if node != townNode:
@@ -486,7 +441,7 @@ class Pathfinder:
                 linkData.append({ 'key': leaveLinkData['key'], 'map': leaveLinkData['map'], 'type': leaveLinkData['type'], 'x': leaveLinkData['x'], 'y': leaveLinkData['y'], 'spawn': leaveLinkData['spawn'] })
 
         # check if we can walk to other nodes
-        delaunay = Delaunator(points)
+        delaunay = Delaunator(list(points))
         
         for i in range(0, len(delaunay.halfedges)):
             halfedge = delaunay.halfedges[i]
@@ -509,7 +464,7 @@ class Pathfinder:
                 linkData.append({ 'key': None, 'map': map, 'type': 'move', 'x': x2, 'y': y2, 'spawn': None })
                 links.append([node2, node1])
                 linkData.append({ 'key': None, 'map': map, 'type': 'move', 'x': x1, 'y': y1, 'spawn': None })
-
+        # print(f"  Link completion: {(datetime.utcnow().timestamp() - linksStart)}\n")
         linkAttr['data'] = linkData
         Pathfinder.graph.add_edges(links, linkAttr)
     
@@ -718,7 +673,7 @@ class Pathfinder:
 
         maps = [Constants.PATHFINDER_FIRST_MAP]
 
-        start = datetime.now()
+        start = datetime.utcnow().timestamp()
         Pathfinder.logger.debug("Preparing Pathfinder...")
 
         i = 0
@@ -760,6 +715,6 @@ class Pathfinder:
                 else:
                     print('The winterland map has changed, cheat to walk to icegolem is not enabled.')
 
-        Pathfinder.logger.debug(f"Pathfinding prepared! ({(datetime.now() - start).total_seconds()}s)")
+        Pathfinder.logger.debug(f"Pathfinding prepared! ({(datetime.utcnow().timestamp() - start)}s)")
         Pathfinder.logger.debug(f"  # Nodes: {len(Pathfinder.graph.vs)}")
         Pathfinder.logger.debug(f"  # Links: {len(Pathfinder.graph.es)}")
\ No newline at end of file
