@@ -13,7 +13,8 @@ from .Ranger import Ranger
 from .Rogue import Rogue
 from .Tools import Tools
 from .Warrior import Warrior
+from .database import Database
 
-__all__ = [ 'Character', 'Constants', 'Entity', 'Game', 'Mage', 'Observer',
-            'Paladin', 'Pathfinder', 'PingCompensatedCharacter', 'Player',
-            'Priest', 'Ranger', 'Rogue', 'Tools', 'Warrior' ]
\ No newline at end of file
+__all__ = [ 'Character', 'Constants', 'Database', 'Entity', 'Game', 'Mage',
+            'Observer', 'Paladin', 'Pathfinder', 'PingCompensatedCharacter',
+            'Player', 'Priest', 'Ranger', 'Rogue', 'Tools', 'Warrior' ]
\ No newline at end of file
